.TH DATIO L Datfile
.SH NAME
d_read[c][m], d_write[c][m], d_seek, d_tell, d_submatrix, d_getsubmatrix,
d_fork, d_getsubdim, convert, subsample, upsample, istype, sametype, samesize
\- dat file I/O package
.SH DESCRIPTION
.PP
These routines provide a level of abstraction for datfile I/O, analogous
to the stdio library.  They provide data type conversions and
simplify the handling of multiple dimensions by providing submatrix
and/or subsampled access to the datfile.
.SH FEATURES
.TP
\(bu
Promotes clarity and reduces programming effort
.PP
.B d_write(df, buf, n)
replaces the usual
.B write(fileno(d_getfp(df)), buf, n*nbps) .
.TP	
\(bu
Data type conversions
.PP
.B d_writec(df, buf, n, buftype)
will write any type of datfile
from a buffer of any other type.  No need for possibly non-portable
casts.  Makes it easy to write programs that work on all types of
datfiles.
.TP	
\(bu
Storage allocation
.PP
.B d_readm(df, &buf, n)
replaces
.B buf = malloc(n*nbps); read(fileno(d_getfp(df)), buf, n*nbps).
.TP	
\(bu
Transparent submatrix/subsampled access
.PP
An optional call to
.B d_submatrix
causes
.B d_write
and
.B d_read
to access an
arbitrary rectangular submatrix, with possible subsampling.
Particularly useful for extracting one color or one frame from a
sequence of RGB frames.
.TP
\(bu
Effortless error detection
.PP
System call and malloc errors are trapped, eliminating the need for
repetitive, tiresome coding like
.PP
.nf
if (read(fileno(d_getfp(df)), buf, n*nbps) < n*nbps) {
	fprintf(stderr,"read failed\\n");
		exit(-1);
	}
.fi
.SH READING AND WRITING
.PP
.nf
\fB
#include <dat.h>
d_read(df, buf, n)
d_readc(df, buf, n, buftype)
d_readm(df, bufptr, n)
d_readcm(df, bufptr, n, buftype)
d_write(df, buf, n)
d_writec(df, buf, n, buftype)
d_writem(df, buf, n)
d_writecm(df, buf, n, buftype)
DATFILE *df;
char buf[], **bufptr;
int n, buftype;
\fP
.fi
These routines transfer
.I n
data elements between the buffer
.I buf
and the
datfile
.IR df ,
where
.I n
does NOT have to be a multiple of the line length.
(As a special case, a negative value of
.I n
means to transfer one line of
data.)  The number of elements actually transferred is returned.  On reads,
this may be less than the number requested when the end of data is reached
and may in fact be 0.
.PP
If the data types of the datfile and the buffer are not the same, use
.B d_readc
and
.BR d_writec ,
which perform the conversion.
.I Buftype
is a
type code of the kind returned by
.B d_gettype
(e.g. 'f'), or 0 for no
conversion.  For conversion rules (i.e., rounding, clipping) see
description of
.BR convert ,
below.
.PP
.B d_readm
and
.B d_readcm
.IR malloc (3)
an input buffer from the heap, returning the
resulting pointer in
.IR bufp .
For complete symmetry with reads,
.B d_writem
and
.B d_writecm
are provided; they return the buffer to the free list.
.SH SEEKING, SUBMATRIX ACCESS, AND SUBSAMPLING
.PP
.nf
\fB
#include <dat.h>
long d_seek(df, where, whence)
long d_submatrix(df, start, stop, step)
int d_fork(df, out, mode)
int d_tell(df, where)
int d_getsubmatrix(df, start, stop, step)
int d_getsubdim(df, size)
DATFILE *df, *out[];
struct d_addr *where, *start, *stop, *step, *size;
int whence, mode;
\fP
.fi
These routines facilitate non-sequential access to the data.
.PP
.nf
\fB
long d_seek(df, where, whence)
DATFILE *df;
struct d_addr *where;
int whence;
\fR
.fi
If
.I whence
is 0, seek to the array element addressed by
.IR where ,
where struct d_addr
is defined in
.B dat.h
as follows:
.PP
.nf
	struct d_addr {		/* multi-dimensional address vector */
		long dset;
		long chan;
		long dim[N_DIMENSIONS];
	};
.fi
.PP
If the datfile has
.I n
dimensions, the elements of the
.I dim
array beyond
the first
.I n
are irrelevant.
A datfile whose descriptor lacks the DATASETS (resp. CHANNELS) keywords
implicitly possesses 1 dataset (channel).  An element of the address vector
may be negative or may be greater than the respective dimension of the
array, in which case the addresses wrap around in raster scan fashion.
If
.I where
is (struct d_addr *)NULL, it is taken to be the zero vector.
The absolute file pointer, divided by bytes/sample, is returned.
.PP
If
.I whence
is 1 or 2, the seek is relative to the current position or
to one past the last data element, respectively.
.nf
\fB
long d_submatrix(df, start, stop, step)
DATFILE *df;
struct d_addr *start, *stop, *step;
\fP
.fi
When
.B d_submatrix
has been called, subsequent calls to
.BR d_read ,
.BR d_write ,
or
.B d_seek
operate within a multidimensional rectangular submatrix of the
datfile.  In each dimension, addressing begins at
.I start
and ends one
element before
.IR stop ,
and the data is subsampled by
.IR step .
In other words,
the datfile is accessed in the following manner:
.nf
	for (d=start.dset; d<stop.dset; d+=step.dset)
		for (c=start.chan; c<stop.chan; c+=step.chan)
			for (i=start.dim[0]; i<stop.dim[0]; i+=step.dim[0])
				for (j=start.dim[1]; j<stop.dim[1]; j+=step.dim[1])
						. . .
.fi
.PP
The call to
.B d_submatrix
seeks the datfile to the first element of the submatrix. 
The absolute file pointer, divided by bytes/sample, is
returned.
.PP
If
.I start
is (struct d_addr *)NULL, it is taken to be the zero vector; if
.I stop
is (struct d_addr *)NULL,
it is taken to be the dimensions vector (as returned by 
.BR d_getdim );
if
.I step
is (struct d_addr *)NULL, the subsampling rate is 1.  Thus,
.B "d_submatrix(df,(struct d_addr *)NULL,(struct d_addr *)NULL,(struct d_addr *)NULL)"
disables submatrix access and rewinds the datfile.
.PP
In order to write on a submatrix with
.IR step >1
in the horizontal dimension,
the datfile must be simultaneously open for reading as well as writing
(i.e.,
.B d_open
with "r+", "w+", or "a+").
.PP
The
.IR scanargv (L)
command-line parser will decode range specifications as defined in
.IR range (L)
and make automatic calls to
.B d_open
and
.BR d_submatrix ,
so that operations on submatrices require no attention from the programmer.
.nf
\fB
int d_fork(df, out, mode)
DATFILE *df, *out[];
int mode;
\fP
.fi
.B d_fork
splits the DATFILE stream
.I df
into a vector of streams
.IR out [],
each restricted to
a subset of the parent (subset of the parent's submatrix, if applicable).
.IR Mode ,
which is an integer between 0 and the number of dimensions-1
or one of the constants
D_ADDR_CHAN or D_ADDR_DSET, selects the dimension to split.
For example, use
D_ADDR_CHAN to split a color image sequence into three streams, one channel
per color separation.
.PP
The parent stream will be flushed but is otherwise unchanged by
.BR d_fork .
Each child will have an I/O pointer independent from those of other streams,
initially positioned at the first element of the subset.
The number of children is returned.
.nf
\fB
int d_tell(df,where)
DATFILE *df;
struct d_addr *where;
\fP
.fi
Get position of next element to be read or written in the current
submatrix, or the whole datfile if submatrix access is not
in effect, and store in
.IR where .
Normally returns 0, but returns -1 if past end of data.
.nf
\fB
int d_getsubmatrix(df, start, stop, step)
DATFILE *df;
struct d_addr *start, *stop, *step;
\fP
.fi
Get the parameters of current submatrix, or the whole datfile if
submatrix access is not in effect, and store in
.IR start ,
.IR stop ,
and
.IR step .
Any of the pointers may be (struct d_addr *)NULL if the information is not
needed.
Returns number of dimensions.
.nf
\fB
int d_getsubdim(df,size)
DATFILE *df;
struct d_addr *size;
\fP
.fi
Get the dimensions of current submatrix, or the whole datfile if
submatrix access is not in effect, and store in
.IR size .
Returns number of
dimensions, or -1 if unknown.
.SH MORE UTILITY SUBROUTINES
.PP
.nf
\fB
#include <dat.h>
convert(obuf, otype, ibuf, itype, n)
subsample(obuf, ibuf, nbps, h_factor, n)
upsample(obuf, ibuf, nbps, h_factor, n)
int istype(d1, itype)
int sametype(d1, d2)
int samesize(d1, d2)
DATFILE *d1, *D2;
char obuf[], ibuf[];
int otype, itype, n, nbps, h_factor;
\fP
.PP
.nf
\fB
convert(obuf, otype, ibuf, itype, n)
char obuf[], ibuf[];
int otype, itype, n;
\fP
.fi
This function converts between any of data types
.BR buUcdDfFzZ ,
and may
be utilized independent of datfiles.
.I Otype
and
.I itype
are
type codes of the kind returned by
.B d_gettype
(e.g. 'f').
The fact that
.I ibuf
is copied to
.I obuf
in order of increasing addresses
should be observed if these buffers overlap in memory.
The following conversion rules are observed:
.PP
.nf
	Real->complex	puts 0 in imaginary part
	Complex->real	retains only real part
	Float->integer	rounds to nearest integer
	Any->integer	clips if out of range
.fi
.PP
Conversions to and from type 'F' (DOUBLE) are most efficient.
.nf
\fB
subsample(obuf, ibuf, nbps, h_factor, n)
char obuf[], ibuf[];
int nbps, h_factor, n;
\fP
.fi
Copy
.I n
samples,
with
.I nbps
bytes/sample, from buffer
.I ibuf
to buffer
.IR obuf ,
subsampling by
.IR h_factor .
.RI ( n
is the number of samples after subsampling.)
Usually
.IR h_factor >0,
but
.IR h_factor <0
can be used to subsample and reverse the data provided that
.I src
points to the last element of the originating buffer.
.nf
\fB
upsample(obuf, ibuf, nbps, h_factor, n)
char obuf[], ibuf[];
int nbps, h_factor, n;
\fP
.fi
Copy
.I n
samples, 
with
.I nbps
bytes/sample, from buffer
.I ibuf
to buffer
.IR obuf ,
upsampling by
.IR h_factor .
The samples of
.I obuf
between those copied from
.I ibuf
are NOT
modified in any way, such as setting to 0.
Usually
.IR h_factor >0,
but
.IR h_factor <0
can be used to upsample and reverse the data provided that
.I dst
points to the last element of the receiving buffer.
.nf
\fB
int istype(d1, itype)
DATFILE *d1;
int itype;
\fP
.fi
Returns 1 if data type of
.I d1
is
.IR itype ,
0 otherwise.
.nf
\fB
int sametype(d1, d2)
DATFILE *d1, *d2;
\fP
.fi
Returns 1 if data types of
.I d1
and
.I d2
match, 0 otherwise.
.nf
\fB
int samesize(d1, d2)
DATFILE *d1, *d2;
\fP
.fi
Returns 1 if the dimensions, channels, and datasets of
.I d1
and
.I d2
match,
0 otherwise.
The size of the submatrix is used for comparison if submatrix access is in
effect.
.SH EXAMPLE
.PP
A program which simultaneously crops, subsamples, and converts
from one 2D datfile, pasting the result on another, can be succinctly
expressed as follows (not necessarily the most efficient).
The submatrix
.IR Infile [64:191,128:383]
is subsampled by 3x2, and the
43x128 result is copied into
.IR Outfile [100:142,100:227],
with the appropriate conversion.
.sp
.nf
#include <dat.h>
main()
{
	DATFILE *in,*out;
	int n;
	double *buf;
	char iname[]="Infile", oname[]="Outfile";
	static struct d_addr cropstart = {0,0,64,128},
		cropstop = {1,1,192,384},
		rate = {1,1,3,2},
		pastestart = {0,0,100,100},
		pastestop = {1,1,143,228};

	in = d_open(iname,"r");
	out = d_open(oname,"a"); /* not "w", which would clobber output */

	d_submatrix(in,		/* establish cropping and subsampling */
		&cropstart,&cropstop,&rate);
	d_submatrix(out,	/* establish paste position */
		&pastestart,&pastestop,(struct d_addr *)NULL);
	while ((n=d_readcm(in,&buf,1024,'F')) > 0)	/* read and convert data */
		d_writecm(out,buf,n,'F');	 /* convert and write data */
	d_close(out);		/* unnecessary, but a good habit */
	d_close(in);
}
.fi
.SH GENERAL NOTES
.PP
On a newly created datfile, the data type and dimensions must be declared
before using these routines and should not be changed once I/O has begun.
Sometimes the length of an output file cannot be predicted in advance;
in that case the extent of the highest-order dimension (or channel or
dataset) should be set initially to some big number, subject to later
revision.
.PP
Do not mix raw Unix I/O (e.g., direct calls to
.BR read (2),
.BR write (2),
or
.BR lseek (2))
with this package.  The result is unpredictable.
.PP
Because the datfile is rewound on the first call to
.BR d_submatrix ,
.BR d_seek ,
.BR d_tell ,
.BR d_write ,
or
.BR d_read ,
.BR d_open 'ing
a datfile with mode "a" or "a+" differs from
"w" or "w+" only in that append mode does not delete an already existing
datfile.
.SH DIAGNOSTICS
.PP
The convention for error handling in the datio package differs from, and is
in some ways more convenient than, the usual Unix philosophy.  When datfile
i/o errors are detected, a message is printed, all datfiles are flushed,
and execution
stops, while Unix system calls merely return -1 and leave error detection
up to the user.
.PP
The following errors are fatal and cause termination with core dump
(SIGQUIT).
.RS
.nf
_dmread: out of core reading (file)
_dread: read error
_dread: early EOF on (file)
_dwrite: read failed for horizontal step
_dwrite: write error
_dwrite: write out of bounds on (file)
d_submatrix: out of core for (file)
d_submatrix: arguments out of range for (file)
d_submatrix: dimensions of (file) unknown
d_submatrix: submatrix too big in (file)
d_seek: seek out of bounds on (file)
_dsize: data type (dtype) not supported
d_fork: invalid dimension selected on (file)
.fi
.RE
.PP	
The following errors are not fatal and causes process suspension (SIGTSTP).
The operation will be retried when the process is restarted.
.RS
.nf
write: No space left on device (resume to retry)
write: Disk quota exceeded (resume to retry)
.fi
.RE
.PP	
The sophisticated user may elect to override the error handling via
.IR signal (2).
If fatal errors are caught or ignored, the datfile subroutine
returns -1.  (Exception: the data type error never returns.)
.SH SEE ALSO
dat(l), datsubs(l), range(l), scanargv(l)
.SH AUTHOR
Steve Hsu <steve@atrp.media.mit.edu>
.SH REVISION
12/11/87
.SH FILES
libdat.a
.br
llib-ldat
.br
dat.h
.SH BUGS
Dimensions can be larger than 65535 only on the vax.
.br
UNSIGNED4 ('U') data type supported only on the vax.
Ascii, bit, and variant data types not handled.
.br
Routines do not work on block- or character-special files which do not
support arbitrary lseeks (e.g., a terminal or raw disk).
